/* Plugin: CPU Stat
 * File: hookedfunctions_linux.cpp
 * Author: Prodigy (jcouzens@uhfco.net)
 * Date: April 21, 2003
 * Info: hooked functions ... 
*/

#include <extdll.h>			//always
#include <string.h>			//to prevent linux errors while processing meta_api.h
#include <meta_api.h>
#include <sdk_util.h>		//pre-defined function definitions saves us a lot of work
#include <cbase.h>
#include "hookedfunctions.h"
#include "plugin.h"
#include "utilfunctions.h"

extern char SERVER_LOG[MAXMSG];
extern char CLIENT_STR[MAXMSG];
extern float fLastDisplayTime;
extern float myPlayerArray[32];
extern int cpuUse;

#ifndef _WIN32
extern float loadAverages[3];
#endif

float fLastFPSCheck;
int serverFPS = 0;

void startFrame(void)
{
	float cpuStatFreq = 0.0f;

	cpuStatFreq = strtofloat(CVAR_GET_POINTER("cpustat_freq")->string);

	// Increment server FPS count
	serverFPS++;

	if (ABSOLUTE_VALUE_EASY(gpGlobals->time - fLastDisplayTime) >= cpuStatFreq)
	{

		int i = 0;
		int pid = 0;
		int numEnts = NUMBER_OF_ENTITIES();
		int fpsShownToClient = 0;
	
		float cpuStatNumCpu = 0.0f;
		float cpuStatDisplay = 0.0f;

		CBaseEntity *pPlayer = NULL;

		hudtextparms_t hudSettings;
			hudSettings.r1 = 255;
			hudSettings.g1 = 0;
			hudSettings.b1 = 0;
			hudSettings.a1 = 0;
			hudSettings.r2 = 255;
			hudSettings.g2 = 255;
			hudSettings.b2 = 250;
			hudSettings.a2 = 0;
			hudSettings.fadeinTime = 0;
			hudSettings.fadeoutTime = 0;
			hudSettings.holdTime = 2;
			hudSettings.fxTime = 0;
			hudSettings.x = 0.80;
			hudSettings.y = 0.80;
			hudSettings.effect = 0;
			hudSettings.channel = 3; 
	
		fLastDisplayTime = gpGlobals->time;

		cpuStatNumCpu = strtofloat(CVAR_GET_POINTER("cpustat_numcpu")->string);
		cpuStatDisplay = strtofloat(CVAR_GET_POINTER("cpustat_display")->string);
		
		// set hold time to check frequency
		hudSettings.holdTime = cpuStatFreq;

		// get CPU Use to go easier on the server, we'll only check if
		// there is a user defined to receive cpu stat 
		for (i=0; i <= 32; ++i)
		{
			if (myPlayerArray[i] == 1.0f)
			{
				getCpuUse();
				i = 32;
			}
		}

		// because of the way we calculate cpuUse, (ie we need two values
		// to perform the math on or else we come back with a garbage
		// value we do this and just say cpu use is zero for our first
		// reading.
		if ((cpuUse < 0) || (cpuUse > 100))
			cpuUse = 0;

// since the NS autohelp system rejects all colour information
// performing these calculations is a waste of time
#ifndef WITH_NS
		if (cpuUse <= 50 * cpuStatNumCpu)
		{
			// green 0%-50% CPU
			hudSettings.r1 = 100;
			hudSettings.g1 = 200;
			hudSettings.b1 = 0;
		}
		else if (cpuUse <= 90 * cpuStatNumCpu) 
		{
			// blue 51%-90% CPU
			hudSettings.r1 = 0;
			hudSettings.g1 = 80;
			hudSettings.b1 = 220;
		}
		else if (cpuUse > 90 * cpuStatNumCpu)
		{
			// red 91%-100% CPU
			hudSettings.r1 = 255;
			hudSettings.g1 = 0;
			hudSettings.b1 = 0;
		}
#endif

		// Make serverFPS accurate to how often we check
		fpsShownToClient = serverFPS / (int)cpuStatFreq;

		// Who knows where this dirty dirty memory has been
		memset(CLIENT_STR,'\0',MAXMSG);

		// NS will truncate the message when it sees a carriage return,
		// so we look for it, and send a single line message to NS instead.
#ifdef _WIN32
	#ifdef WITH_NS
		sprintf(CLIENT_STR,"CPU in Use: %i%% Entity Count: %i Server FPS: %i\n",cpuUse,numEnts,fpsShownToClient);
	#else
		sprintf(CLIENT_STR,"CPU in Use: %i%%\nEntity Count: %i\nServer FPS: %i\n",cpuUse,numEnts,fpsShownToClient);
	#endif
#else
	#ifndef WITH_NS
		sprintf(CLIENT_STR,"Load Avg: [%0.2f, %0.2f, %0.2f]\nActual CPU: %i%%\nEntities: %i\nServer FPS: %i\n",
			loadAverages[0],loadAverages[1],loadAverages[2],cpuUse,numEnts,fpsShownToClient);
	#else
		sprintf(CLIENT_STR,"Load Avg: [%0.2f, %0.2f, %0.2f] CPU: %i%% Entities: %i Server FPS: %i\n",
			loadAverages[0],loadAverages[1],loadAverages[2],cpuUse,numEnts,fpsShownToClient);
	#endif
#endif

		// main loop, this delivers the messages to all the clients
		for (i=0; i < 32; ++i)
		{
			if (myPlayerArray[i] == 1.0f)
			{
				pPlayer = UTIL_PlayerByIndex(i);

				if (pPlayer != NULL)
				{
					pid = ENTINDEX(pPlayer->edict());

					if (pid <= 0 || pid > 32)
						RETURN_META(MRES_IGNORED);
		
					if (cpuStatDisplay == 0.0f)
						UTIL_HudMessage(pPlayer, hudSettings, CLIENT_STR);
					else
						sendTextMsg(pPlayer->edict(), HUD_PRINTTALK, CLIENT_STR);
		
				} // if player exists
			} // if player is selected
		} // for every player 

	// clean out serverFPS so its ready for next check
	serverFPS = 0;

	} // if its time

        RETURN_META(MRES_HANDLED);
}

BOOL clientConnect( edict_t *pEntity, const char *pszName, const char *pszAddress, char szRejectReason[ 128 ] )
{
        int pid;

        pid = ENTINDEX(pEntity);

        myPlayerArray[pid] = 0.0f;

        RETURN_META_VALUE(MRES_HANDLED,TRUE);
}

void clientDisconnect(edict_t *pEntity)
{
        int pid;

        pid = ENTINDEX(pEntity);

        myPlayerArray[pid] = 0.0f;

        RETURN_META(MRES_HANDLED);
}