/* Plugin: CPU Stat
 * File: utilfunctions.cpp
 * Author: Prodigy (jcouzens@uhfco.net)
 * Date: April 21, 2003
 * Info: util functions
*/

#include <extdll.h>		//always
#include <string.h>		//to prevent linux errors while processing meta_api.h
#include <sdk_util.h>	//pre-defined function definitions saves us a lot of work
#include <meta_api.h>	// Plugin_info, etc
#include <stdio.h>	// for fopen
#include <cbase.h>
#include <player.h>

#include "plugin.h"
#include "utilfunctions.h"

// Stuff explicitly for the CPU utilitiy fuctions
#ifdef _WIN32
#define SYS_BASIC_INFO       0
#define SYS_PERFORMANCE_INFO 2
#define SYS_TIME_INFO        3

#define Li2Double(x) ((double)((x).HighPart) * 4.294967296E9 + (double)((x).LowPart))

typedef struct
{
    DWORD   dwUnknown1;
    ULONG   uKeMaximumIncrement;
    ULONG   uPageSize;
    ULONG   uMmNumberOfPhysicalPages;
    ULONG   uMmLowestPhysicalPage;
    ULONG   uMmHighestPhysicalPage;
    ULONG   uAllocationGranularity;
    PVOID   pLowestUserAddress;
    PVOID   pMmHighestUserAddress;
    ULONG   uKeActiveProcessors;
    BYTE    bKeNumberProcessors;
    BYTE    bUnknown2;
    WORD    wUnknown3;
} SYSTEM_BASIC_INFORMATION;

typedef struct
{
    LARGE_INTEGER   liIdleTime;
    DWORD           dwSpare[76];
} SYSTEM_PERFORMANCE_INFORMATION;

typedef struct
{
    LARGE_INTEGER liKeBootTime;
    LARGE_INTEGER liKeSystemTime;
    LARGE_INTEGER liExpTimeZoneBias;
    ULONG         uCurrentTimeZoneId;
    DWORD         dwReserved;
} SYSTEM_TIME_INFORMATION;

typedef LONG (WINAPI *PROCNTQSI)(UINT,PVOID,ULONG,PULONG);

PROCNTQSI NtQuerySystemInformation;
#endif

// Global Defines
DLL_GLOBAL const Vector     g_vecZero = Vector(0,0,0);

char SERVER_LOG[MAXMSG] = " ";
char CLIENT_STR[MAXMSG] = " ";

int gmsgTextMsg = 0;
int gmsgSayText = 0;

#ifdef _WIN32
LARGE_INTEGER	myOldCpuTime = {0,0};
LARGE_INTEGER   myOldIdleTime = {0,0};
#else
long myOldCpuTime;
float loadAverages[3];
#endif

int cpuUse = 0;

//=========================================================
// UTIL_LogPrintf - Prints a logged message to console.
// Preceded by LOG: ( timestamp ) < message >
//=========================================================
void UTIL_LogPrintf(char *fmt, ...)
{
	va_list			argptr;
	static char		string[1024];
	
	va_start (argptr, fmt);
	vsprintf (string, fmt, argptr);
	va_end   (argptr);

	// Print to server console
	ALERT(at_logged, "%s", string);
}


// returns a CBaseEntity pointer to a player by index.  Only returns if the player is spawned and connected
// otherwise returns NULL
// Index is 1 based

CBaseEntity *getPlayerFromUserId(int uid)
{
	int i;
	CBaseEntity *pPlayer = NULL;

	for (i = 1; i <= gpGlobals->maxClients; i++)
	{
		pPlayer = UTIL_PlayerByIndex(i);

		if (pPlayer)
		{
			if (GETPLAYERUSERID(pPlayer->edict())==uid)
				return(pPlayer);
		}
	}
	
	return NULL;
}

CBaseEntity *UTIL_PlayerByIndex(int playerIndex)
{
	CBaseEntity *pPlayer = NULL;

	if (playerIndex > 0 && playerIndex <= gpGlobals->maxClients)
	{
		edict_t *pPlayerEdict = INDEXENT(playerIndex);
		if (pPlayerEdict && !pPlayerEdict->free)
		{
			pPlayer = CBaseEntity::Instance(pPlayerEdict);	
		}
	}
	
	return pPlayer;
}

void sendTextMsg(edict_t *pEntity, int msg_dest, const char *msg_name)
{
	if (FNullEnt(pEntity))
		return;
	
	if (gmsgSayText == 0)
		gmsgSayText = GET_USER_MSG_ID(PLID,"TextMsg",NULL);

	if (gmsgSayText == 0 || pEntity->v.flags & FL_FAKECLIENT)
		return;
	
	MESSAGE_BEGIN(MSG_ONE, gmsgSayText, NULL, pEntity);
		WRITE_BYTE(msg_dest);
		WRITE_STRING("%s");
		WRITE_STRING(msg_name);
	MESSAGE_END();
}
	
void UTIL_HudMessage(CBaseEntity *pEntity, const hudtextparms_t &textparms, const char *pMessage)
{
	if((int)GETPLAYERUSERID(pEntity->edict()) == 0) return;

	MESSAGE_BEGIN(MSG_ONE, SVC_TEMPENTITY, NULL, pEntity->edict());
		WRITE_BYTE(TE_TEXTMESSAGE);
		WRITE_BYTE(textparms.channel & 0xFF);
  
		WRITE_SHORT(FixedSigned16(textparms.x, 1<<13 ));
		WRITE_SHORT(FixedSigned16(textparms.y, 1<<13 ));
		WRITE_BYTE(textparms.effect);
  
		WRITE_BYTE(textparms.r1 );
		WRITE_BYTE(textparms.g1 );
		WRITE_BYTE(textparms.b1 );
		WRITE_BYTE(textparms.a1 );
  
		WRITE_BYTE(textparms.r2 );
		WRITE_BYTE(textparms.g2 );
		WRITE_BYTE(textparms.b2 );
		WRITE_BYTE(textparms.a2 );
  
		WRITE_SHORT(FixedUnsigned16(textparms.fadeinTime, 1<<8 ));
		WRITE_SHORT(FixedUnsigned16(textparms.fadeoutTime, 1<<8 ));
		WRITE_SHORT(FixedUnsigned16(textparms.holdTime, 1<<8 ));
  
	if (textparms.effect == 2)
		WRITE_SHORT(FixedUnsigned16(textparms.fxTime, 1<<8 ));
  
	if (strlen(pMessage) < 512)
	{
		WRITE_STRING(pMessage);
	}
	else
	{
		char tmp[512];
		strncpy( tmp, pMessage, 511);
		tmp[511] = 0;
		WRITE_STRING(tmp);
	}
	MESSAGE_END();
}

int UTIL_GetPlayerIndex(CBasePlayer *pPlayer)
{
	if(pPlayer == NULL) return 0;
	int counter;
	for(counter = 1; counter <= gpGlobals->maxClients; counter++)
	{
		if(pPlayer == UTIL_PlayerByIndex(counter))
			return counter;
	}
	return 0;
}

unsigned short FixedUnsigned16(float value, float scale)
{
	int output;

	output = (int)(value * scale);
	if (output < 0)
		output = 0;
	if (output > 0xFFFF)
		output = 0xFFFF;

	return (unsigned short)output;
}

short FixedSigned16(float value, float scale)
{
	int output;

	output = (int)(value * scale);

	if (output > 32767)
		output = 32767;

	if (output < -32768)
		output = -32768;

	return (short)output;
}

#ifdef _WIN32
void getCpuUse(void)
{
	long status = 0;

	double dbIdleTime;
	double dbSystemTime;

	SYSTEM_PERFORMANCE_INFORMATION SysPerfInfo;
	SYSTEM_TIME_INFORMATION SysTimeInfo;
	SYSTEM_BASIC_INFORMATION SysBaseInfo;

	NtQuerySystemInformation = (PROCNTQSI)GetProcAddress(GetModuleHandle("ntdll"),"NtQuerySystemInformation");

	if (!NtQuerySystemInformation)
		RETURN_META(MRES_IGNORED);

	// get number of processors in the system
	status = NtQuerySystemInformation(SYS_BASIC_INFO, &SysBaseInfo, sizeof(SysBaseInfo), NULL);

	if (status != NO_ERROR)
		RETURN_META(MRES_IGNORED);

	// get new system time
	status = NtQuerySystemInformation(SYS_TIME_INFO, &SysTimeInfo, sizeof(SysTimeInfo), 0);

	if (status!=NO_ERROR)
		RETURN_META(MRES_IGNORED);

	// get new CPU's idle time
	status = NtQuerySystemInformation(SYS_PERFORMANCE_INFO, &SysPerfInfo, sizeof(SysPerfInfo), NULL);

	if (status != NO_ERROR)
		RETURN_META(MRES_IGNORED);

	if (myOldCpuTime.QuadPart != 0)
	{
	
		// CurrentValue = NewValue - OldValue
		dbIdleTime = Li2Double(SysPerfInfo.liIdleTime) - Li2Double(myOldIdleTime);
		dbSystemTime = Li2Double(SysTimeInfo.liKeSystemTime) - Li2Double(myOldCpuTime);	

		// CurrentCpuIdle = IdleTime / SystemTime
		dbIdleTime = dbIdleTime / dbSystemTime;
	
		// CurrentCpuUsage% = 100 - (CurrentCpuIdle * 100) / NumberOfProcessors
		dbIdleTime = 100.0 - dbIdleTime * 100.0 / (double)SysBaseInfo.bKeNumberProcessors + 0.5;

		// cpuUse = IdleTime / Check Frequency
		cpuUse = (int)dbIdleTime;
#ifdef DEBUG
		sprintf(SERVER_LOG,"[CPUSTAT] Sys: %i Idle: %i oSys: %i oIdle: %i CPU: %i\n",
			dbIdleTime,dbSystemTime,myOldCpuTime,myOldIdleTime,cpuUse);
		SERVER_PRINT(SERVER_LOG);
#endif
	}

	// store the old values
	myOldIdleTime = SysPerfInfo.liIdleTime;
	myOldCpuTime = SysTimeInfo.liKeSystemTime;

	RETURN_META(MRES_HANDLED);
}
#else
void getCpuUse(void)
{
	float cpuStatFreq = 0.0f;
	float cpuStatNumCpu = 0.0f;

	long cpuTime = 0;
	long cpuNice = 0;
	long diff = 0;

	char proc[MAXMSG] = " ";
	char *tmpPtr = NULL;

	FILE *loadPtr;
	FILE *procPtr;

	cpuStatFreq = strtofloat(CVAR_GET_POINTER("cpustat_freq")->string);
	cpuStatNumCpu = strtofloat(CVAR_GET_POINTER("cpustat_numcpu")->string);

	// This is being anal, but trust me it is intentional, either through
	// my own combination of lack of sleep and excessive caffiene, or through
	// some other means, I often found these pointers STILL defined, even tho
	// as you will see below, I NULL them out.

	if (loadPtr != NULL)
		loadPtr = NULL;

	if (procPtr != NULL)
		procPtr = NULL;

	// You won't see any windblows programmer go to these lengths...
	// ... I gots teh hardcore love for all my unix homies

#ifdef WITH_FREEBSD
	loadPtr = fopen("/usr/compat/linux/proc/loadavg", "rt");
#else
	loadPtr = fopen("/proc/loadavg", "rt");
#endif

	if (procPtr != NULL)
		procPtr = NULL;

#ifdef WITH_FREEBSD
	procPtr = fopen("/usr/compat/linux/proc/stat", "rt");
#else
	procPtr = fopen("/proc/stat", "rt");
#endif

	if (loadPtr != NULL)
	{
		// strtok is your friend, its like perl's split kinda.  Only, it has a the
		// bad side effect of modifying the original data, good thing we don't care!

		fgets(proc,MAXMSG,loadPtr);
		tmpPtr = strtok(proc," ");
		loadAverages[0] = atof(tmpPtr);
		tmpPtr = strtok(NULL," ");
		loadAverages[1] = atof(tmpPtr);
		tmpPtr = strtok(NULL," ");
		loadAverages[2] = atof(tmpPtr);
#ifdef DEBUG
		sprintf(SERVER_LOG,"[CPUSTAT] Load average: %0.2f, %0.2f, %0.2f\n",loadAverages[0],loadAverages[1],loadAverages[2]);
		SERVER_PRINT(SERVER_LOG);
#endif
		fclose(loadPtr);
		loadPtr = NULL;
	}
	else
	{
		sprintf(SERVER_LOG,"[CPUSTAT] Unable to open file pointer...\n");
		SERVER_PRINT(SERVER_LOG);
		RETURN_META(MRES_IGNORED);
	}

	// Get our overall system cpu use
	if (procPtr != NULL)
	{
		fgets(proc,MAXMSG,procPtr);
		tmpPtr = strtok(proc," ");
		tmpPtr = strtok(NULL," ");
		cpuTime = atoi(tmpPtr);
		tmpPtr = strtok(NULL," ");
		cpuNice = atoi(tmpPtr);

		// This is so obvious I don't think I need to spell it out...
		diff = cpuTime + cpuNice;
		cpuUse = diff - myOldCpuTime;
		cpuUse = cpuUse / (int)cpuStatFreq;

		// Check for SMP systems and adjust the value accordingly
		if (cpuStatNumCpu > 1)
			cpuUse = cpuUse / (int)cpuStatNumCpu;

#ifdef DEBUG
		sprintf(SERVER_LOG,"[CPUSTAT] CPU Time: %i CPU Nice: %i Combined: %i CPU Old Combined: %i CPU IN USE: %i\n",
			(int)cpuTime,(int)cpuNice,(int)diff,(int)myOldCpuTime,cpuUse);
		SERVER_PRINT(SERVER_LOG);
#endif

		// store the old value
		myOldCpuTime = cpuTime+cpuNice;

		fclose(procPtr);
		procPtr = NULL;
	}
	else
	{
		sprintf(SERVER_LOG,"[CPUSTAT] Unable to open file pointer...\n");
		SERVER_PRINT(SERVER_LOG);
		RETURN_META(MRES_IGNORED);
	}

	RETURN_META(MRES_HANDLED);
}
#endif
